@@ -6,6 +6,7 @@ from inspect import isfunction
 from operator import mul
 from functools import partial, reduce, wraps
 
+from axial_positional_embedding import AxialPositionalEmbedding
 from product_key_memory import PKM
 from routing_transformer.reversible import ReversibleSequence, SequentialSequence
 
@@ -93,7 +94,7 @@ def ema_inplace(moving_avg, new, decay):
     if is_empty(moving_avg):
         moving_avg.data.copy_(new)
         return
-    moving_avg.data.mul_(decay).add_(new, alpha= (1 - decay))
+    moving_avg.data.mul_(decay).add_(1 - decay, new)
 
 # helper classes
 
@@ -163,55 +164,6 @@ class RelativePositionalEmbedding(nn.Module):
         emb = torch.einsum('bhnid,jhd->bhnij', q, self.weights.type(q.dtype)) * self.scale
         return shift(emb)
 
-class AxialPositionalEmbedding(nn.Module):
-    def __init__(self, dim, max_seq_len, axial_shape = ()):
-        super().__init__()
-        assert reduce(mul, axial_shape, 1) == max_seq_len, 'axial position shape must multiply up to max sequence length'
-
-        self.dim = dim
-        self.seq_len = max_seq_len
-        self.shape = axial_shape
-
-        self.weights = ParameterList(self, 'weights', len(axial_shape))
-
-        for ind, shape in enumerate(self.shape):
-            ax_shape = [1] * len(self.shape)
-            ax_shape[ind] = shape
-            ax_shape = (1, *ax_shape, dim)
-            ax_emb = nn.Parameter(torch.zeros(ax_shape).normal_(0, 1))
-            self.weights.append(ax_emb)
-
-    def forward(self, x):
-        b, t, e = x.shape
-        embs = []
-
-        for ax_emb in self.weights.to_list():
-            expand_shape = (b, *self.shape, self.dim)
-            emb = ax_emb.expand(expand_shape).reshape(b, self.seq_len, self.dim)
-            embs.append(emb)
-
-        pos_emb = sum(embs)
-        return pos_emb[:, :t].to(x)
-
-# a mock parameter list object until below issue is resolved
-# https://github.com/pytorch/pytorch/issues/36035
-class ParameterList(object):
-    def __init__(self, kls, prefix, length):
-        self.ind = 0
-        self.kls = kls
-        self.prefix = prefix
-        self.length = length
-
-    def _keyname(self, prefix, ind):
-        return f'{prefix}_{ind}'
-
-    def append(self, x):
-        setattr(self.kls, self._keyname(self.prefix, self.ind), x)
-        self.ind += 1
-
-    def to_list(self):
-        return [getattr(self.kls, self._keyname(self.prefix, i)) for i in range(self.length)]
-
 # local attention
 
 class LocalAttention(nn.Module):
@@ -661,7 +613,7 @@ class RoutingTransformerLM(nn.Module):
         self.max_seq_len = max_seq_len
 
         self.token_emb = nn.Embedding(num_tokens, emb_dim)
-        self.axial_pos_emb = AxialPositionalEmbedding(emb_dim, max_seq_len, axial_shape=(max_seq_len // window_size, window_size))
+        self.axial_pos_emb = AxialPositionalEmbedding(emb_dim, axial_shape=(max_seq_len // window_size, window_size))
         self.routing_transformer = RoutingTransformer(dim, depth, max_seq_len, heads = heads, window_size = window_size, local_attn_window_size = local_attn_window_size, causal = causal, weight_tie = weight_tie, ff_dropout = ff_dropout, attn_dropout = attn_dropout, attn_layer_dropout = attn_layer_dropout, layer_dropout = layer_dropout, n_local_attn_heads = n_local_attn_heads, ff_glu = ff_glu, reversible = reversible, ff_chunks = ff_chunks, kmeans_ema_decay = kmeans_ema_decay, receives_context = receives_context, context_window_size = context_window_size, rel_pos_emb = rel_pos_emb, pkm_layers = pkm_layers, pkm_num_keys = pkm_num_keys, _register_kmeans_update = _register_kmeans_update)
 
         if emb_dim != dim:
